#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <iostream.h>
#include <Inventor/Xt/SoXt.h>
#include <Inventor/Xt/viewers/SoXtExaminerViewer.h>
#include <Inventor/draggers/SoTranslate1Dragger.h>
#include <Inventor/nodes/SoCamera.h>
#include <Inventor/nodes/SoComplexity.h>
#include <Inventor/nodes/SoCone.h>
#include <Inventor/nodes/SoCoordinate3.h>
#include <Inventor/nodes/SoCoordinate4.h>
#include <Inventor/nodes/SoCube.h>
#include <Inventor/nodes/SoCylinder.h>
#include <Inventor/nodes/SoIndexedFaceSet.h>
#include <Inventor/nodes/SoLineSet.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoPickStyle.h>
#include <Inventor/nodes/SoShapeHints.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoSphere.h>
#include <Inventor/nodes/SoSwitch.h>
#include <Inventor/nodes/SoTexture2.h>
#include <Inventor/nodes/SoTextureCoordinate2.h>
#include <Inventor/nodes/SoTransform.h>
#include <Inventor/sensors/SoFieldSensor.h>
#include <Inventor/sensors/SoTimerSensor.h>
#include <Inventor/nodes/SoText2.h>
#include <Inventor/engines/SoCompose.h>
#include "harm.cxx"

SoSeparator *VectorFieldGridRoot;
SoSeparator *LDraggerRoot;
SoSeparator *TDraggerRoot;
SbVec3f *VectorFieldCoord;
SoSeparator *EmitterRoot;
SoSeparator *StreamlineRoot;
SoCoordinate3 *StreamlineCoord;
SbVec3f EmitterPoint(0.0, 0.0, 0.0);
SoXtExaminerViewer *myViewer;

int NumStreamlineSamples = 60;

float DraggerScale = 0.2;
float EmitterScale = 0.4;

float PresentTime = 0.0;
float DTime = 0.01;

const int xDim = 7;
const int yDim = 7;
const int zDim = 7;


////////////////////////////////////////////////////////////////
//
// Global variables for L and R draggers
// 

char LdraggerName[16];
char TdraggerName[16];
int Lvalue=0;
int Tvalue=0;


////////////////////////////////////////////////////////////
//
// (C) Copyright 2000 
// David C. Banks Florida State University
//
// Heavily modified by Kevin Beason
// 1/25/00
//



////////////////////////////////////////////////////////////
//
// Define a vector for each point

void
vectorFieldEvaluate (const SbVec3f &p, SbVec3f *v)
  {
  *v = SbVec3f(p[0], p[1], p[2]);
  v->normalize();
  }

int visFirstTime=1;
int numVis=0;
SoSeparator *rootSep;
SoSeparator *visSep;

void visUpdate(){

  // add harm map visualization
  double phi, theta;
  double l, r;
  SbVec3f pointer;

  if (visFirstTime){  
    visSep=new SoSeparator();
    rootSep->addChild(visSep);
  }
  
  SoSeparator *mySep;
  SoTransform *myTran;
  SoSphere *mySphere;
  int i=0;

  double actual_theta, actual_phi;
  
  for (phi=0; phi<=M_PI; phi+=M_PI/40.0){
    for (theta=0; theta<=2.0*M_PI; theta+=M_PI/20.0){

      // wiggle theta and phi
      actual_theta=theta*(drand48()/5.0+1.0);
      actual_phi=phi*(drand48()/5.0+1.0);
      
      if (visFirstTime){
	mySep = new SoSeparator();
	myTran = new SoTransform();
	mySphere = new SoSphere();
      }
      else {
	mySep = (SoSeparator *)visSep->getChild(i);
	myTran = (SoTransform *)mySep->getChild(0);
	mySphere = (SoSphere *)mySep->getChild(1);
      }

      mySphere->radius.setValue(0.01);
      r=1.0;
      pointer[2]=r*cos(actual_phi);
      l=r*sin(actual_phi);
      pointer[0]=l*sin(actual_theta);
      pointer[1]=l*cos(actual_theta);
      pointer.normalize();
      myTran->translation.setValue(pointer*harm(actual_phi, actual_theta,
						Lvalue, Tvalue)*10.0);

      if (visFirstTime){
	visSep->addChild(mySep);
	mySep->addChild(myTran);
	mySep->addChild(mySphere);
      }

      if (visFirstTime) numVis++;
      i++;
    }
  }

  visFirstTime=0;
  return;
}

////////////////////////////////////////////////////////////
//
// From a starting point p, produce another point pNext
// in the direction of the vector field.

int my_index;
double harms[xDim*yDim*zDim]={0.0};

void
vectorFieldIntegrate (SbVec3f &p, SbVec3f *pNext, float direction)
  {
  SbVec3f v;
  float h = 0.1/xDim*10.0*harm(p[0], p[1], p[2], Lvalue, Tvalue);

  vectorFieldEvaluate(p, &v);
  *pNext = p + v*h*direction;
  }

////////////////////////////////////////////////////////////
//
// Update sphere texture
//
void LMupdate(){
  char textureName[200];
  sprintf(textureName, "pics/%1il%1im.gif", Lvalue, Tvalue);
  SoTexture2 *emitterTexture2 = (SoTexture2 *) EmitterRoot->getChild(3);
  emitterTexture2->filename.setValue(textureName);

  visUpdate();
}

////////////////////////////////////////////////////////////
//
// Create a grid full of little arrows aligned with the
// vector field.


SoSeparator *
makeGrid ()
  {
  int xi, yi, zi;
  float x, y, z;
  float dx, dy, dz;
  SbVec3f vectorVal;
  static int firstTime = 1;

  dx = 2.0/(xDim-1);
  dy = 2.0/(yDim-1);
  dz = 2.0/(zDim-1);

  SoSeparator *root;
  SoCone *cone;
  SbVec3f vertex;
  SbVec3f VectorFieldCoord;

  if (firstTime)
    {
    root = new SoSeparator();
    VectorFieldGridRoot = root;

    cone = new SoCone();
    cone->height = 0.5/xDim;
    cone->bottomRadius = 0.1/xDim;
    cone->parts = SoCone::SIDES;
    }
  else
    {
    root = VectorFieldGridRoot;
    }

  SoSeparator *coneSep;
  SoTransform *coneTransform;

  static int ziRestart = zDim, yiRestart = yDim, xiRestart = xDim;
  int indexChange = 0; 
  int maxIndexChange = 501;

  if (ziRestart >= zDim) 
    {
    ziRestart = yiRestart = xiRestart = 0;
    srand48(0);
    }
  zi = ziRestart;
  z  = -1.0 + zi*dz;
  yi = yiRestart;
  y  = -1.0 + yi*dy;
  xi = xiRestart;
  x  = -1.0 + xi*dx;

  float phase, size, angle;
  float nextAngle;
  int zeroCrossing;

  //fprintf(stderr, "%d %d %d\n\n", xi, yi, zi);
  for (; zi < zDim; zi++, ziRestart++, z += dz, yi = yiRestart = 0, y = -1.0)
    {
    for (; yi < yDim; yi++, yiRestart++, y += dy, xi = xiRestart = 0, x = -1.0)
      {
      for (; xi < xDim; xi++, xiRestart++, x += dx)
	{
	my_index = zi*xDim*yDim + yi*xDim + xi;
       
	vertex = /* SbVec3f(x, y, z) + */
	         SbVec3f(0.0, 0.0, 0.0) +
	  SbVec3f(drand48()*dx, drand48()*dy, drand48()*dz) -
	  SbVec3f(dx, dy, dz)*0.5;
	phase = drand48();

	if (firstTime)
	  {
	  coneSep = new SoSeparator();
	  coneTransform = new SoTransform();
	  coneTransform->translation.setValue(vertex);

	  root->addChild(coneSep);
	  coneSep->addChild(coneTransform);
	  coneSep->addChild(cone);
	  }
	else
	  {
	  coneSep = (SoSeparator *) root->getChild(my_index);
	  coneTransform = (SoTransform *) coneSep->getChild(0);
	  }

	angle = fmodf(2.0*M_PI*phase + 20.0*PresentTime, 2.0*M_PI);
	nextAngle = fmodf(angle + 20.0*DTime*
  /*  harm(vertex[0], vertex[1], vertex[2], Lvalue, Tvalue)*10.0* */
	  ((xDim*yDim*zDim)/maxIndexChange + 1), 2.0*M_PI);

	zeroCrossing = (nextAngle <= M_PI) && (angle >= M_PI);
	size = sin(0.5*angle);
	size *= size;

	// multiply size by Sphereical harmonic
	size *= harm(vertex[0], vertex[1], vertex[2], Lvalue, Tvalue);
	size *= 20.0;

	

	if (!zeroCrossing)
	  {
	  vertex = coneTransform->translation.getValue();
	  vectorFieldIntegrate(vertex, &vertex, 2.0);
	  }
	else
	  {
	  }

	coneTransform->translation.setValue(vertex);
	coneTransform->scaleFactor.setValue(size, size, size);

	vectorFieldEvaluate(vertex, &VectorFieldCoord);

	vectorVal = VectorFieldCoord;
	vectorVal.normalize();
	coneTransform->rotation.setValue(
	  SbRotation(SbVec3f(0.0, 1.0, 0.0), -vectorVal));

	if (!firstTime)
	  indexChange++; 
	if (indexChange > maxIndexChange) 
	  {
	  xiRestart ++;
	  break;
	  }
	}
      if (indexChange > maxIndexChange) 
	{
	break;
	}
      }
    if (indexChange > maxIndexChange) 
      {
      break;
      }
    }
 
  firstTime = 0;
  return root;
  }


////////////////////////////////////////////////////////////
//
// When the L dragger moves, re-orient its probe arrow to follow
// the vector field.
//
// Integrate a streamline through the dragger.

void
LdraggerUpdate(void *data, SoDragger *dragger)
  {
  SoTranslate1Dragger *LDragger;
  LDragger = (SoTranslate1Dragger *) dragger;

  SoTransform *LdraggerTextTrans;
  LdraggerTextTrans = (SoTransform *) LDraggerRoot->getChild(3);

  SoText2 *LdraggerText;
  LdraggerText = (SoText2 *) LDraggerRoot->getChild(4);

  float x, y, z;
  LDragger->translation.getValue().getValue(x,y,z);

  // set Lvalue
  int tempint;
  tempint=(int) x;
  if (tempint<0) Lvalue=0;
  else if (tempint>5) Lvalue=5;
  else Lvalue=tempint;

  sprintf(LdraggerName, "L = %i", Lvalue);
  LdraggerText->string.setValue(LdraggerName);

  LdraggerTextTrans->translation.setValue(SbVec3f(x-0.4, 0.5, 0.0));

  // update M (Tvalue and the T label)
  if (Tvalue<-Lvalue) Tvalue=-Lvalue;
  if (Tvalue>Lvalue) Tvalue=Lvalue;
  SoText2 *TdraggerText;
  TdraggerText = (SoText2 *) TDraggerRoot->getChild(4);
  sprintf(TdraggerName, "M = %i", Tvalue);
  TdraggerText->string.setValue(TdraggerName);

  LMupdate();
}

////////////////////////////////////////////////////////////
//
// When the T dragger moves, re-orient its probe arrow to follow
// the vector field.
//
// Integrate a streamline through the dragger.

void
TdraggerUpdate(void *data, SoDragger *dragger)
  {
  SoTranslate1Dragger *TDragger;
  TDragger = (SoTranslate1Dragger *) dragger;

  SoTransform *TdraggerTextTrans;
  TdraggerTextTrans = (SoTransform *) TDraggerRoot->getChild(3);

  SoText2 *TdraggerText;
  TdraggerText = (SoText2 *) TDraggerRoot->getChild(4);

  float x, y, z;
  TDragger->translation.getValue().getValue(x,y,z);

  // set Lvalue
  int tempint;
  tempint=(int) x;
  if (tempint<-Lvalue) Tvalue=-Lvalue;
  else if (tempint>Lvalue) Tvalue=Lvalue;
  else Tvalue=tempint;
    
  sprintf(TdraggerName, "M = %i", Tvalue);
  TdraggerText->string.setValue(TdraggerName);

  TdraggerTextTrans->translation.setValue(SbVec3f(x-0.4, 0.5, 0.0));

  LMupdate();
}


////////////////////////////////////////////////////////////
//
// The scene contains a 3D translation dragger and some
// other elements that cannot be accidentally picked.
//
//                           scene
//                             |
//     -----------------------------------------------------
//     |          |          |          |         |        |
//     0          1          2          3         4        5
//     |          |          |          |         |        |
//  2Sided   complexity   dragger  vectorField  probe streamline

SoSeparator *
makeScene ()
  {
  SoSeparator *root = new SoSeparator();
  rootSep=root;

  SoShapeHints *twoSided = new SoShapeHints();
  twoSided->shapeType = SoShapeHints::UNKNOWN_SHAPE_TYPE;
  twoSided->vertexOrdering = SoShapeHints::COUNTERCLOCKWISE;
  root->addChild(twoSided);

  SoComplexity *myComplexity = new SoComplexity();
  myComplexity->textureQuality = 1.0;
  root->addChild(myComplexity);

  ////////////////////////////////////////////////////////////
  //
  // Create L dragger
  //
  //                       dragger
  //                          |
  //          ---------------------------------------
  //          |          |            |      |      |
  //          0          1            2      3      4
  //          |          |            |      |      |
  //      trans/scale   orient     dragger trans  text

  SoSeparator *LdragSep = new SoSeparator();
  LDraggerRoot = LdragSep;
  root->addChild(LdragSep);

  SoTransform *LdragTrans = new SoTransform();
  LdragTrans->translation.setValue(-1.0, 2.0, 0.0);
  LdragTrans->scaleFactor.setValue(DraggerScale, DraggerScale, DraggerScale);
  LdragSep->addChild(LdragTrans); // dragSep.child(0)

  SoTransform *LdragOrient = new SoTransform();
  LdragSep->addChild(LdragOrient); // dragSep.child(1)

  SoTranslate1Dragger *LDragger = new SoTranslate1Dragger();
  LdragSep->addChild(LDragger); // dragSep.child(2)
  LDragger->translation.setValue(0.0, 0.0, 0.0);

  SoTransform *LdraggerTextTrans = new SoTransform();
  LdraggerTextTrans->translation.setValue(0.0-0.4, 0.5, 0.0);
  LdragSep->addChild(LdraggerTextTrans); // dragSep.child(3)
  
  SoText2 *LdraggerText = new SoText2();
  LdraggerText->string.setValue("L = 0");
  LdragSep->addChild( LdraggerText ); // dragSep.child(4)

  LDragger->addMotionCallback(LdraggerUpdate, LDragger);

  ////////////////////////////////////////////////////////////
  //
  // Create T dragger
  //
  //                       dragger
  //                          |
  //          ---------------------------------------
  //          |          |            |      |      |
  //          0          1            2      3      4
  //          |          |            |      |      |
  //      trans/scale   orient     dragger trans  text

  SoSeparator *TdragSep = new SoSeparator();
  TDraggerRoot = TdragSep;
  root->addChild(TdragSep);

  SoTransform *TdragTrans = new SoTransform();
  TdragTrans->translation.setValue(-1.0, 1.5, 0.0);
  TdragTrans->scaleFactor.setValue(DraggerScale, DraggerScale, DraggerScale);
  TdragSep->addChild(TdragTrans); // dragSep.child(0)

  SoTransform *TdragOrient = new SoTransform();
  TdragSep->addChild(TdragOrient); // dragSep.child(1)

  SoTranslate1Dragger *TDragger = new SoTranslate1Dragger();
  TdragSep->addChild(TDragger); // dragSep.child(2)
  TDragger->translation.setValue(0.0, 0.0, 0.0);

  SoTransform *TdraggerTextTrans = new SoTransform();
  TdraggerTextTrans->translation.setValue(0.0-0.4, 0.5, 0.0);
  TdragSep->addChild(TdraggerTextTrans); // dragSep.child(3)
  
  SoText2 *TdraggerText = new SoText2();
  TdraggerText->string.setValue("M = 0");
  TdragSep->addChild( TdraggerText ); // dragSep.child(4)

  TDragger->addMotionCallback(TdraggerUpdate, TDragger);

  
  ////////////////////////////////////////////////////////////
  //
  // Create the vector field grid
  //
  //                       vectorField
  //                          |
  //          -------------------------
  //          |          |            |
  //          0          1            2
  //          |          |            |
  //      pickStyle   material      grid

  SoSeparator *vectorFieldSep = new SoSeparator();
  root->addChild(vectorFieldSep);

  SoPickStyle *myPickStyle = new SoPickStyle();
  myPickStyle->style.setValue(SoPickStyle::UNPICKABLE);
  vectorFieldSep->addChild( myPickStyle );

  SoMaterial *myMaterial = new SoMaterial();
  myMaterial->ambientColor.setValue(0.0, 0.0, 0.0);
  myMaterial->diffuseColor.setValue(0.3, 0.7, 0.8);
  myMaterial->specularColor.setValue(0.0, 0.0, 0.0);
  myMaterial->emissiveColor.setValue(0.0, 0.0, 0.0);
  myMaterial->transparency.setValue(0.0);
  vectorFieldSep->addChild(myMaterial);

  vectorFieldSep->addChild(makeGrid());


  ////////////////////////////////////////////////////////////
  //
  // Create a probe to follow the dragger. This probe has
  // an arrow that aligns to the vector field.
  //
  //                       EmitterRoot
  //                          |
  //          ----------------------------------------------
  //          |          |            |          |         |
  //          0          1            2          3         4
  //          |          |            |          |         |
  //      pickStyle    scale      transform   texture   sphere

  EmitterRoot = new SoSeparator();
  root->addChild(EmitterRoot);

  EmitterRoot->addChild( myPickStyle ); // child(0)

  SoTransform *emitterScale = new SoTransform();
  emitterScale->scaleFactor.setValue(EmitterScale, EmitterScale, EmitterScale);
  EmitterRoot->addChild(emitterScale); // child(1)

  SoTransform *emitterTrans = new SoTransform();
  emitterTrans->translation.setValue(0.0, 0.0, 0.0);
  EmitterRoot->addChild(emitterTrans); // child(2)

  SoTexture2 *emitterTexture2 = new SoTexture2();
  emitterTexture2->filename.setValue("pics/0l0m.gif");
  EmitterRoot->addChild( emitterTexture2 ); // child(3)

  SoSphere *emitterSphere = new SoSphere();
  emitterSphere->radius.setValue(1.0);
  EmitterRoot->addChild( emitterSphere ); // child(4)

  visUpdate();

  return root;
  }


void
animate(void *, SoSensor *)
  {
  makeGrid ();
  PresentTime += DTime;
  }



void
main(int, char **argv)
  {
  Widget myWindow = SoXt::init(argv[0]);
  if (myWindow == NULL) exit(1);

  SoSeparator *root = makeScene();

  SoTimerSensor *sensor = new SoTimerSensor(animate, root);
  sensor->setInterval(SbTime(0.015)); // 70 frames per second
  sensor->schedule();



  myViewer = new SoXtExaminerViewer(myWindow);
  myViewer->setTransparencyType(SoGLRenderAction::BLEND);
  myViewer->setSceneGraph(root);
  myViewer->setTitle("Emitter");
  myViewer->show();


  SoXt::show(myWindow);
  SoXt::mainLoop();
  }
