#include<math.h>
#include<iostream.h>
#include<stdlib.h>

double PHI(double p, int m);
double THETA(double t, int l, int m);
double harm(double p, double t, int l, int m);

double harm(double x, double y, double z, int l, int m){
   // calculate PHI(phi,theta)*THETA(phi,theta)

   // first convert (x,y,z) to polar coordinates (p,t)
   double r, p, t;
   if (x==0.0 && y==0.0){
      p=0.0;
      t=0.0;
   }
   else { // r!=0
      r=sqrt(x*x+y*y+z*z);
      p=acos(z/r);
      t=atan2(x,y);
   }
   
   //cout << "(r,p,t)=(" << r << ", " << p << ", " << t << ")\n";
   return(harm(p,t,l,m));
}

double harm(double p, double t, int l, int m){
  double h=PHI(p,m)*THETA(t,l,m);
  return(h*h);
}

double PHI(double p, int m){
  if (m==0) return (1.0/sqrt(2.0*M_PI));
  if (m<0){
    return (1.0/sqrt(M_PI)*sin((double)(m)*p));
  }
  else {
    return (1.0/sqrt(M_PI)*cos((double)m*p));
  }
}

double THETA(double t, int l, int m){
   int absm=m;
   if (m<0) absm=-m;
   if (!(absm<=l)) {
      cerr << "Error in THETA(): |m| is not less than l!\n";
      exit(1);
   }
   
   if (l==0) return (1.0/sqrt(2.0));
   else if (l==1){
      if (absm==0) return (sqrt(6.0)/2.0*cos(t));
      else return (sqrt(3.0)/2.0*sin(t));
   }
   else if (l==2){
      if (absm==0) return (sqrt(10.0)/4.0*(3.0*pow(cos(t),2.0)-1.0));
      else if (absm==1) return (sqrt(15.0)/2.0*sin(t)*cos(t));
      else return (sqrt(15.0)/4.0*pow(sin(t),2.0));
   }
   else if (l==3){
      if (absm==0) return (3.0*sqrt(14.0)/4.0*(5.0/3.0*pow(cos(t),3.0)-
					       cos(t)));
      else if (absm==1) return (sqrt(42.0)/8.0*sin(t)*(5.0*pow(cos(t),2.0)-
						       1.0));
      else if (absm==2) return (sqrt(105.0)/4.0*pow(sin(t),2.0)*cos(t));
      else return (sqrt(70.0)/8.0*pow(sin(t),3.0));
   }
   else if (l==4){
      if (absm==0) return (9.0*sqrt(2.0)/16.0*(35.0/3.0*pow(cos(t),4.0)-
					       10.0*pow(cos(t),2.0)+1.0));
      else if (absm==1) return (9.0*sqrt(10.0)/8.0*sin(t)*
				(7.0/3.0*pow(cos(t),3.0)-cos(t)));
      else if (absm==2) return (3.0*sqrt(5.0)/8.0*pow(sin(t),2.0)*
				(7.0*pow(cos(t),2.0)-1.0));
      else if (absm==3) return (3.0*sqrt(70.0)/8.0*pow(sin(t),3.0)*cos(t));
      else return (3.0*sqrt(35.0)/16.0*pow(sin(t),4.0));
   }
   else if (l==5){
      if (absm==0) return (15.0*sqrt(22.0)/16.0*
			   (21.0/5.0*pow(cos(t),5.0)-14.0/3.0*pow(cos(t),3.0)+
			    cos(t)));
      else if (absm==1) return (sqrt(165.0)/16.0*sin(t)*
				(21*pow(cos(t),4.0)-14.0*pow(cos(t),2.0)+1.0));
      else if (absm==2) return (sqrt(1155.0)/8.0*pow(sin(t),2.0)*
				(3.0*pow(cos(t),3.0)-cos(t)));
      else if (absm==3) return (sqrt(770.0)/32.0*pow(sin(t),3.0)*
				(9*pow(cos(t),2.0)-1.0));
      else if (absm==4) return (3.0*sqrt(385.0)/16.0*pow(sin(t),4.0)*cos(t));
      else return (3.0*sqrt(154.0)/32.0*pow(sin(t),5.0));
   }
   else {
      cerr << "Error in THETA(): l is too big (>5)!\n";
      exit(1);
   }
   
   // this is to stop g++ from giving a warning
   return(0.0);
}
